#!/usr/bonsaitools/bin/perl -w
# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Contributor(s): Myk Melez <myk@mozilla.org>

############################################################################
# Module Initialization
############################################################################

# make it harder for us to do dangerous things in Perl
use diagnostics;
use strict;

package Request;

# Use the template toolkit (http://www.template-toolkit.org/) to generate
# the user interface (HTML pages and mail messages) using templates in the
# "templates/" subdirectory.
use Template;

# This is the global template object that gets used one or more times by
# the script when it needs to process a template and return the results.
# Configuration parameters can be specified here that apply to all templates
# processed in this file.
my $template = Template->new(
  {
    # Colon-separated list of directories containing templates.
    INCLUDE_PATH => 'templates' ,
    # Allow templates to be specified with relative paths.
    RELATIVE => 1 
  }
);

# This module requires that its caller have said "require CGI.pl" to import
# relevant functions from that script and its companion globals.pl.

############################################################################
# Functions
############################################################################

sub fulfillRequest 
{
  my ($objecttype, $objectid, @fieldnames) = @_;

  # !!! Also receive the user's ID!
  
  # To fulfill a request, a user must submit a form that contains
  # the field specified in the request.  This function receives
  # the type of object being updated (f.e. bug or attachment), the 
  # ID of that object, and a list of fieldnames that were updated 
  # by the form; and it marks fulfilled all requests that are 
  # satisfied by this update and sends off email to the requesters 
  # to notify them that their request has been fulfilled.

  # First, get a list of request definitions for the specified
  # table/fields.  If no fields were specified, get definitions
  # for any field in the table.
  my $query = "SELECT id FROM requestdefs ";
  $query .= "WHERE objecttype = " . &::SqlQuote($objecttype);
  if (scalar(@fieldnames))
  {
    $query .= " AND fieldname IN ('" . join("', '", @fieldnames) . "')";
  }
  &::SendSQL($query);
  my @defids;
  while (my ($defid) = &::FetchSQLData())
  {
    push @defids , $defid;
  }

  # If there aren't any request definitions for the table/fields
  # combination, don't continue any further.
  return if !scalar(@defids);

  # Build the criteria used to retrieve requests to update, starting
  # with all requests connected to the definitions retrieved above.
  my $criteria = "defid IN (" . join(", ", @defids) . ")";

  # ??? do i need to do this, or can i assume it has already been done
  # and use something like $::userid?  Or require it to be passed in?
  # Add the user ID for the current user to the criteria.
  my $userid = &::confirm_login();
  $criteria .= " AND fulfillerid = $userid";

  # Only the requests on this specific object (attachment, bug, etc.)
  # should be fulfilled.
  $criteria .= " AND objectid = $objectid";

  # Only unfulfilled requests should be fulfilled
  $criteria .= " AND fulfilldate IS NULL";

  # Get a list of matching requests
  &::SendSQL("SELECT id FROM requests WHERE " . $criteria);
  my @requestids;
  while (my ($requestid) = &::FetchSQLData())
  {
    push @requestids, $requestid;
  }

  # If there aren't any requests to fulfill, don't continue any further.
  return if !scalar(@requestids);

  # Mark requests as fulfilled in the database.
  &::SendSQL("UPDATE requests SET fulfilldate = NOW() WHERE " . $criteria);

  # Get comprehensive information about the fulfilled requests
  # for the email notification.
  &::SendSQL("
      SELECT    fulfillers.realname , 
                fulfillers.login_name , 
                requesters.realname , 
                requesters.login_name , 
                requestdefs.description , 
                requestdefs.fieldname , 
                requestdefs.fieldvalue , 
                requestdefs.objecttype , 
                requestdefs.cclist , 
                requests.objectid 
      FROM      requests, requestdefs, profiles AS requesters, profiles AS fulfillers
      WHERE     requests.id IN (" . join(", ", @requestids) . ")
      AND       requests.defid = requestdefs.id
      AND       requests.requesterid = requesters.userid
      AND       requests.fulfillerid = fulfillers.userid
      ORDER BY  fulfillerid , fulfilldate , description , objectid
    ");
  my @requests;
  while (my ($fname, $femail, $rname, $remail, $desc, $fldname, $fldval, $otype, $cclist, $oid) = &::FetchSQLData())
  {
    push 
      @requests , 
      { 
        'fulfillername' => $fname , 
        'fulfilleremail' => $femail , 
        'requestername' => $rname , 
        'requesteremail' => $remail , 
        'description' => $desc , 
        'fieldname' => $fldname , 
        'fieldvalue' => $fldval , 
        'objecttype' => $otype , 
        'cclist' => $cclist , 
        'objectid' => $oid 
      };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'Param' => \&::Param , # for retrieving global parameters
      'PerformSubsts' => \&::PerformSubsts # for processing global parameters
    };

  my $sendmailparam = &::Param("sendmailnow") ? "" : "-ODeliveryMode=deferred";

  # Send email notification to the requesters of each fulfilled request
  foreach my $request (@requests)
  {
    # Add summary information about the object to the request.
    ($request->{'bugsummary'}, $request->{'attachsummary'}, $request->{'bugid'}) = 
      Request::getSummaryInfo($request->{'objectid'}, $request->{'objecttype'});

    # Add the request to the collection of variables passed into the template.
    $vars->{'request'} = $request;

    open(SENDMAIL, "|/usr/lib/sendmail $sendmailparam -t")
      || DisplayError("Can't open sendmail")
      && exit;

    $template->process("request/fulfillmail.atml", $vars, \*SENDMAIL)
      || &::DisplayError("Template process failed: " . $template->error())
      && exit;

    close SENDMAIL;
  }

  return \@requests;

}


sub listFulfilledRequests
{
  my ($requests) = @_;

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'requests' => $requests , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/listFulfilledRequests.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub matchUsers
{
  # Generates a list of users whose login name (email address) or real name
  # matches a substring.  Useful for finding the list of potential fulfillers 
  # for a submitted request.

  my ($substr) = @_;

  # quote the substring and add percent signs so it works as a substring
  # search inside a SQL LIKE statement
  $substr = &::SqlQuote('%' . $substr . '%');

  my @users;

  # run the query and build an array of hashes containing the results
  &::SendSQL("SELECT    userid, login_name, realname 
              FROM      profiles 
              WHERE     login_name LIKE $substr 
              OR        realname LIKE $substr 
              ORDER BY  realname 
              LIMIT     100
             ");
  while (my ($userid, $login_name, $realname) = &::FetchSQLData())
  {
    push @users , { 'userid' => $userid , 'login_name' => $login_name , 'realname' => $realname };
  }

  return \@users;

}


sub getUser
{
  # Generates a list of users whose login name (email address) or real name
  # matches a substring.  Useful for finding the list of potential fulfillers 
  # for a submitted request.

  my ($userid) = @_;

  # run the query and build a hash containing the results
  &::SendSQL("SELECT    login_name, realname 
              FROM      profiles 
              WHERE     userid = $userid
             ");

  my ($login_name, $realname) = &::FetchSQLData();

  # !!! change this to $user = {}; return $user;
  my %user = ( 'userid' => $userid , 'login_name' => $login_name , 'realname' => $realname );

  return \%user;

}

sub getSummaryInfo
{
  # Gets summary information about an object for which a request was made.
  # If the object is a bug, returns the value in its bugs.short_desc field.
  # If the object is an attachment, returns the value in its attachments.-
  # description field as well as summary information for the associated bug.
  # Either way, return the ID of the bug or associated bug.

  my ($id, $type) = @_;

  my $bugsummary = "";
  my $attachsummary = "";
  my $bugid = "";

  if ($type eq 'bug')
  {
    &::SendSQL("SELECT short_desc , '' , bug_id FROM bugs WHERE bug_id = $id");
  }
  elsif ($type eq 'attachment')
  {
    &::SendSQL("SELECT bugs.short_desc , attachments.description , bugs.bug_id
                FROM attachments INNER JOIN bugs ON attachments.bug_id = bugs.bug_id
                WHERE attachments.attach_id = $id");
  }

  ($bugsummary, $attachsummary, $bugid) = &::FetchSQLData();

  return ($bugsummary, $attachsummary, $bugid);
}
