#!/usr/bonsaitools/bin/perl -w
# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Contributor(s): Terry Weissman <terry@mozilla.org>
#                 Myk Melez <myk@mozilla.org>

################################################################################
# Script Initialization
################################################################################

# Make it harder for us to do dangerous things in Perl.
use diagnostics;
use strict;

# Include the Bugzilla CGI and general utility library.
require "CGI.pl";

# Establish a connection to the database backend.
ConnectToDatabase();

# Use the template toolkit (http://www.template-toolkit.org/) to generate
# the user interface (HTML pages and mail messages) using templates in the
# "templates/" subdirectory.
use Template;

# This is the global template object that gets used one or more times by
# the script when it needs to process a template and return the results.
# Configuration parameters can be specified here that apply to all templates
# processed in this file.
my $template = Template->new(
  {
    # Colon-separated list of directories containing templates.
    INCLUDE_PATH => 'templates' ,
    # Allow templates to be specified with relative paths.
    RELATIVE => 1 
  }
);

# Check whether or not the user is logged in and, if so, set the $::userid 
# and $::usergroupset variables.
quietly_check_login();

################################################################################
# Data Validation / Security Authorization
################################################################################

# Validate the value of the "id" form field, which must contain a positive
# integer that is the ID of an existing attachment.
if (defined $::FORM{'id'}) 
{
  $::FORM{'id'} =~ /^[1-9][0-9]*$/
    || DisplayError("You didn't enter a valid attachment number.") 
    && exit;
  
  # Make sure the attachment exists in the database.
  SendSQL("SELECT bug_id FROM attachments WHERE attach_id = $::FORM{'id'}");
  MoreSQLData()
    || DisplayError("Attachment #$::FORM{'id'} does not exist.") 
    && exit;

  # Make sure the user is authorized to access this attachment's bug.
  my ($bugid) = FetchSQLData();
  ValidateBugID($bugid)
    || DisplayError("You are not authorized to access that bug or its attachments.") 
    && exit;
}

# Validate the value of the "bugid" field.
if (defined $::FORM{'bugid'})
{
  ValidateBugID($::FORM{'bugid'});
}

# !!! Validate other form fields!

# Create a list of valid attachment status flags.
# ??? Move this into versioncache?
SendSQL("SELECT id, name FROM attachstatusdefs ORDER BY sortkey");
my %statusdefs;
while (my ($id, $name) = FetchSQLData())
{
  $statusdefs{$id} = $name;
}

# Validate the values of the "status" form field.
if (defined $::MFORM{'status'}) 
{
  foreach my $status (@{$::MFORM{'status'}})
  {
    defined($statusdefs{$status})
      || DisplayError("One or more of the attachment's status flags is invalid.") 
      && exit;
  }
}

# Determine whether to use the action specified by the user or the default.
my $action = $::FORM{'action'} || 'view';

################################################################################
# Main Body Execution
################################################################################

# All calls to this script should contain an "action" variable whose value
# determines what the user wants to do.  The code below checks the value of
# that variable and runs the appropriate code.

if ($action eq "view")  
{ 
  # View an attachment.  Returns the attachment using the MIME type specified in
  # the attachment record.
  view($::FORM{'id'}); 
}

elsif ($action eq "viewall") 
{ 
  # Display all attachments in an HTML page with each attachment contained within
  # an HTML IFRAME element.
  viewall($::FORM{'bugid'}); 
}

elsif ($action eq "edit") 
{ 
  # Edit an attachment record.  This action allows the user to edit the 
  # attachment's description, MIME type, ispatch and isactive flags, statuses,
  # and to submit a comment that appears in the bug.  It does not allow the user
  # to edit the content of the attachment itself.
  edit($::FORM{'id'}); 
}

elsif ($action eq "update") 
{ 
  # Update an attachment record.
  confirm_login();
  UserInGroup("editbugs")
    || DisplayError("You are not authorized to edit attachments.")
    && exit;
  # !!! verify that the attach_id field was submitted and throw an error if not!
  update(
    $::FORM{'id'} , 
    $::FORM{'comment'} , 
    $::FORM{'description'} , 
    $::FORM{'mimetype'} , 
    $::FORM{'ispatch'} , 
    $::FORM{'isactive'} , 
    @{$::MFORM{'status'}} 
  );
}

elsif ($action eq "editold") 
{ 
  # Edit an attachment record using regular frames instead of IFRAMEs.  Deprecated.
  editold($::FORM{'id'}); 
}

elsif ($action eq "console") 
{ 
  # Display form for editing attachment when using "editold" method.  Deprecated.
  console($::FORM{'id'}); 
}

# Actions below this line can only be performed by a user with "tweakparams"
# privileges (i.e. a system administrator).

elsif ($action eq "adminDefs") 
{ 
  # Administer attachment status flags, which is the set of status flags 
  # that can be applied to an attachment.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  adminDefs(); 
} 

elsif ($action eq "createDef") 
{ 
  # Display a form for creating a new attachment status flag.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  createDef(); 
} 

elsif ($action eq "insertDef") 
{ 
  # Insert a new attachment status flag into the database.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  insertDef(
    $::FORM{'name'} , 
    $::FORM{'desc'} , 
    $::FORM{'sortkey'} 
  );
}

elsif ($action eq "editDef") 
{ 
  # Display a form for editing an existing attachment status flag.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  editDef($::FORM{'defid'}); 
} 

elsif ($action eq "updateDef") 
{ 
  # Update an attachment status flag in the database.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  updateDef(
    $::FORM{'defid'} , 
    $::FORM{'name'} , 
    $::FORM{'desc'} , 
    $::FORM{'sortkey'} 
  );
}

elsif ($action eq "deleteDef") 
{ 
  # Delete an attachment status flag from the database.
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer attachment status flags.")
    && exit;
  deleteDef($::FORM{'defid'}); 
} 

else 
{ 
  # If the action that the user wants to take (specified in the "action" form field)
  # is none of the above listed actions, display an error telling the user that we
  # do not understand what they would like to do.
  DisplayError("I could not figure out what you wanted to do.")
    && exit;
}

################################################################################
# Functions
################################################################################

sub view
{
  # View an attachment.  Returns the attachment using the MIME type specified in
  # the attachment record.

  my ($attachid) = @_;

  # Retrieve the attachment content and its MIME type from the database.
  SendSQL("SELECT mimetype, thedata FROM attachments WHERE attach_id = $attachid");
  my ($mimetype, $thedata) = FetchSQLData();

  # Determine what mode the user wants to view the attachment in for attachments
  # that contain potentially dangerous "script" tags, and display the attachment
  # in the appropriate mode (or display a warning asking the user to choose a mode).
  $::FORM{'mode'} ||= '';
  if ( $::FORM{'mode'} eq 'disabled' ) {
    # Disable scripts by converting their opening and closing tags into character entities.
    $thedata =~ s/<\/?([\w.-]*:?script[^>]*)>/&lt;$1&gt;/gi;
  } elsif ( $::FORM{'mode'} eq 'original' ) {
    # Don't do anything to scripts in the attachment.
  } elsif ( $::FORM{'mode'} eq 'plain' ) {
    # Change the MIME type.
    $mimetype = 'text/plain';
  } elsif ( $thedata =~ /<[\w.-]*:?script/i ) {
    # Define the variables and functions that will be passed to the UI template.
    my $vars = 
      {
        'attachid' => $attachid , 
        'Param' => \&Param , # for retrieving global parameters
        'PerformSubsts' => \&PerformSubsts # for processing global parameters
      };

    # Return the appropriate HTTP response headers.
    print "Content-type: text/html\n\n";

    # Generate and return the UI (HTML page) from the appropriate template.
    $template->process("attachment/viewwarning.atml", $vars)
      || DisplayError("Template process failed: " . $template->error())
      && exit;

    return;
  }
    
  # Display the attachment.
  print "Content-Type: $mimetype\n\n";
  print $thedata;
}


sub viewall
{
  # Display all attachments in a series of IFRAMEs within one HTML page.

  my ($bugid) = @_;

  # Retrieve a list of attachments for this bug and write them into an array
  # of hashes in which each hash represents a single attachment.
  SendSQL("
           SELECT attach_id, creation_ts, mimetype, description, ispatch, isactive 
           FROM attachments WHERE bug_id = $bugid ORDER BY attach_id
         ");
  my @attachments = ();
  while (MoreSQLData()) {
    my %a;
    ($a{'attachid'}, $a{'date'}, $a{'mimetype'}, $a{'description'}, $a{'ispatch'}, $a{'isactive'}) = FetchSQLData();

    # Format the attachment's creation/modification date into something readable.
    if ($a{'date'} =~ /^(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$/) {
        $a{'date'} = "$3/$4/$2&nbsp;$5:$6";
    }

    # Quote HTML characters (&<>) in the description so they display correctly.
    $a{'description'} = &::value_quote($a{'description'});

    # Flag attachments as to whether or not they can be viewed (as opposed to
    # being downloaded).  Currently I decide they are viewable if their MIME type 
    # is either text/.* or application/vnd.mozilla.*.
    # !!! Yuck, what an ugly hack.  Fix it!
    $a{'isviewable'} = ( $a{'mimetype'} =~ /^(text|image|application\/vnd\.mozilla\.)/ );

    # Retrieve a list of status flags that have been set on the attachment.
    &::PushGlobalSQLState();
    &::SendSQL(" 
                SELECT   name 
                FROM     attachstatuses, attachstatusdefs 
                WHERE    attach_id = $a{'attachid'} 
                AND      attachstatuses.statusid = attachstatusdefs.id
                ORDER BY sortkey
              ");
    my @statuses = ();
    while (&::MoreSQLData()) { 
      my ($status) = &::FetchSQLData(); 
      push @statuses , $status;
    }
    $a{'statuses'} = \@statuses;
    &::PopGlobalSQLState();

    push @attachments, \%a;
  }

  # Retrieve the bug summary for displaying on screen.
  SendSQL("SELECT short_desc FROM bugs WHERE bug_id = $bugid");
  my ($bugsummary) = FetchSQLData();

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'bugid' => $bugid , 
      'bugsummary' => $bugsummary , 
      'attachments' => \@attachments , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/viewall.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub edit
{
  # Edit an attachment record.  This action allows the user to edit the 
  # attachment's description, MIME type, ispatch and isactive flags, statuses,
  # and to submit a comment that appears in the bug.  It does not allow the user
  # to edit the content of the attachment itself.

  my ($attachid) = @_;

  # Retrieve the attachment from the database.
  SendSQL("SELECT description, mimetype, bug_id, ispatch, isactive 
           FROM attachments WHERE attach_id = $attachid");
  my ($description, $mimetype, $bugid, $ispatch, $isactive) = FetchSQLData();

  # Flag attachment as to whether or not it can be viewed (as opposed to
  # being downloaded).  Currently I decide it is viewable if its MIME type 
  # is either text/.* or application/vnd.mozilla.*.
  # !!! Yuck, what an ugly hack.  Fix it!
  my $isviewable = ( $mimetype =~ /^(text|image|application\/vnd\.mozilla\.)/ );

  # Retrieve a list of status flags that have been set on the attachment.
  my %statuses;
  SendSQL("SELECT id, name 
           FROM attachstatuses JOIN attachstatusdefs 
           WHERE attachstatuses.statusid = attachstatusdefs.id 
           AND attach_id = $attachid");
  while (my ($id, $name) = FetchSQLData())
  {
    $statuses{$id} = $name;
  }

  # Retrieve a list of all status flags and build an array of hashes
  # in which each hash is a status flag.
  # ???: Move this into versioncache?
  my @statusdefs;
  SendSQL("SELECT id, name FROM attachstatusdefs ORDER BY sortkey");
  while (MoreSQLData())
  {
    my ($id, $name) = FetchSQLData();
    push @statusdefs, { 'id' => $id , 'name' => $name };
  }

  # Retrieve a list of attachments for this bug as well as a summary of the bug
  # to use in a navigation bar across the top of the screen.
  SendSQL("SELECT attach_id FROM attachments WHERE bug_id = $bugid ORDER BY attach_id");
  my @attachments;
  while (MoreSQLData())
  {
    my ($id) = FetchSQLData();
    push @attachments, $id;
  }
  SendSQL("SELECT short_desc FROM bugs WHERE bug_id = $bugid");
  my ($bugsummary) = FetchSQLData();

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'attachid' => $attachid , 
      'description' => $description , 
      'mimetype' => $mimetype , 
      'bugid' => $bugid , 
      'bugsummary' => $bugsummary , 
      'ispatch' => $ispatch , 
      'isactive' => $isactive , 
      'isviewable' => $isviewable , 
      'statuses' => \%statuses , 
      'statusdefs' => \@statusdefs , 
      'attachments' => \@attachments , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/edit.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub update
{
  # Update an attachment record.

  my ($attachid, $comment, $description, $mimetype, $ispatch, $isactive, @statuses) = @_;

  # Set ispatch and isactive to zero (representing boolean false in the database) 
  # if they are undefined, which happens when their checkboxes are unchecked
  # on the "edit attachment" form, since unchecked HTML checkboxes are not sent
  # in HTML POST requests.
  $ispatch ||= 0;
  $isactive ||= 0;

  # Get the bug ID for the bug to which this attachment is attached.
  SendSQL("SELECT bug_id FROM attachments WHERE attach_id = $attachid");
  my $bugid = FetchSQLData() 
    || DisplayError("Cannot figure out bug number.")
    && exit;

  # Lock database tables in preparation for updating the attachment.
  SendSQL("LOCK TABLES attachments WRITE , attachstatuses WRITE , 
           attachstatusdefs READ , fielddefs READ , bugs_activity WRITE");

  # Get a copy of the attachment record before we make changes
  # so we can record those changes in the activity table.
  SendSQL("SELECT description, mimetype, ispatch, isactive 
           FROM attachments WHERE attach_id = $attachid");
  my ($olddescription, $oldmimetype, $oldispatch, $oldisactive) = FetchSQLData();

  # Get the list of old status flags.
  SendSQL("SELECT    attachstatusdefs.name 
           FROM      attachments, attachstatuses, attachstatusdefs
           WHERE     attachments.attach_id = $attachid
           AND       attachments.attach_id = attachstatuses.attach_id
           AND       attachstatuses.statusid = attachstatusdefs.id
           ORDER BY  attachstatusdefs.sortkey
          ");
  my @oldstatuses;
  while (MoreSQLData()) {
    push(@oldstatuses, FetchSQLData());
  }
  my $oldstatuslist = join(', ', @oldstatuses);

  # Update the database with the new status flags.
  SendSQL("DELETE FROM attachstatuses WHERE attach_id = $attachid");
  foreach my $statusid (@statuses) 
  {
    SendSQL("INSERT INTO attachstatuses (attach_id, statusid) VALUES ($attachid, $statusid)");
  }

  # Get the list of new status flags.
  SendSQL("SELECT    attachstatusdefs.name 
           FROM      attachments, attachstatuses, attachstatusdefs
           WHERE     attachments.attach_id = $attachid
           AND       attachments.attach_id = attachstatuses.attach_id
           AND       attachstatuses.statusid = attachstatusdefs.id
           ORDER BY  attachstatusdefs.sortkey
          ");
  my @newstatuses;
  while (MoreSQLData()) {
    push(@newstatuses, FetchSQLData());
  }
  my $newstatuslist = join(', ', @newstatuses);

  # Quote "description" and "mimetype" for use in the SQL UPDATE statement.
  my $quoteddescription = SqlQuote($description);
  my $quotedmimetype = SqlQuote($mimetype);

  # Update the attachment record in the database.
  SendSQL("UPDATE  attachments 
           SET     description = $quoteddescription , 
                   mimetype = $quotedmimetype , 
                   ispatch = $ispatch , 
                   isactive = $isactive 
           WHERE   attach_id = $attachid
         ");

  # Record changes in the activity table.
  if ($olddescription ne $description) {
    my $quotedolddescription = SqlQuote($olddescription);
    my $fieldid = GetFieldID('attachments.description');
    SendSQL("INSERT INTO bugs_activity (bug_id, who, bug_when, fieldid, oldvalue, newvalue) 
             VALUES ($bugid, $::userid, NOW(), $fieldid, $quotedolddescription, $quoteddescription)
            ");
  }
  if ($oldmimetype ne $mimetype) {
    my $quotedoldmimetype = SqlQuote($oldmimetype);
    my $fieldid = GetFieldID('attachments.mimetype');
    SendSQL("INSERT INTO bugs_activity (bug_id, who, bug_when, fieldid, oldvalue, newvalue) 
             VALUES ($bugid, $::userid, NOW(), $fieldid, $quotedoldmimetype, $quotedmimetype)
            ");
  }
  if ($oldispatch ne $ispatch) {
    my $fieldid = GetFieldID('attachments.ispatch');
    SendSQL("INSERT INTO bugs_activity (bug_id, who, bug_when, fieldid, oldvalue, newvalue) 
             VALUES ($bugid, $::userid, NOW(), $fieldid, $oldispatch, $ispatch)
            ");
  }
  if ($oldisactive ne $isactive) {
    my $fieldid = GetFieldID('attachments.isactive');
    SendSQL("INSERT INTO bugs_activity (bug_id, who, bug_when, fieldid, oldvalue, newvalue) 
             VALUES ($bugid, $::userid, NOW(), $fieldid, $oldisactive, $isactive)
            ");
  }
  if ($oldstatuslist ne $newstatuslist) {
    my $quotedoldstatuslist = SqlQuote($oldstatuslist);
    my $quotednewstatuslist = SqlQuote($newstatuslist);
    my $fieldid = GetFieldID('attachstatusdefs.name');
    SendSQL("INSERT INTO bugs_activity (bug_id, who, bug_when, fieldid, oldvalue, newvalue) 
             VALUES ($bugid, $::userid, NOW(), $fieldid, $quotedoldstatuslist, $quotednewstatuslist)
            ");
  }

  # Unlock all database tables now that we are finished updating the database.
  SendSQL("UNLOCK TABLES");

  # If this installation has enabled the request manager, let the manager know
  # an attachment was updated so it can check for requests on that attachment
  # and fulfill them.  The request manager allows users to request database
  # changes of other users and tracks the fulfillment of those requests.  When
  # an attachment record is updated and the request manager is called, it will
  # fulfill those requests that were requested of the user performing the update
  # which are requests for the attachment being updated.
  my $requests;
  if (Param('userequestmanager'))
  {
    use Request;
    # Specify the fieldnames that have been updated.
    my @fieldnames = ('description', 'mimetype', 'status', 'ispatch', 'isactive');
    # Fulfill pending requests.
    $requests = Request::fulfillRequest('attachment', $attachid, @fieldnames);
  }

  # If the user submitted a comment while editing the attachment, 
  # add the comment to the bug.
  if ($comment)
  {
    # Append a string to the comment to let users know that the comment came from
    # the "edit attachment" screen.
    $comment .= qq|\n(From update of attachment $attachid)\n|;

    # Get the user's login name since the AppendComment function needs it.
    my $who = DBID_to_name($::userid);
    # Mention $::userid again so Perl doesn't give me a warning about it.
    my $neverused = $::userid;

    # Append the comment to the list of comments in the database.
    AppendComment($bugid, $who, $comment);

  }

  # Send mail to let people know the bug has changed.  Uses a special syntax
  # of the "open" and "exec" commands to capture the output of "processmail",
  # which "system" doesn't allow, without running the command through a shell,
  # which backticks (``) do.
  #system ("./processmail", $bugid , $::userid);
  #my $mailresults = `./processmail $bugid $::userid`;
  open(PMAIL, "-|") or exec('./processmail', $bugid, $::userid);
  my $mailresults = '';
  while (<PMAIL>) {
    $mailresults .= $_;
  }
  close(PMAIL);
 
  # !!! Figure out whether or not the shadow database needs to be updated!

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'attachid' => $attachid , 
      'bugid' => $bugid , 
      'requests' => $requests , 
      'mailresults' => $mailresults , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/updated.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub editold
{
  # Edit an attachment record using regular frames instead of IFRAMEs.  Deprecated.

  my ($attachid) = @_;

  print "Content-type: text/html\n\n";

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'attachid' => $attachid
    };
  $template->process("attachment/editold.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub console
{
  # Display form for editing attachment when using "editold" method.  Deprecated.

  my ($attachid) = @_;

  SendSQL("SELECT description, mimetype, bug_id 
              FROM attachments 
              WHERE attach_id = $attachid");
  my ($description, $mimetype, $bug_id) = FetchSQLData();

  my %statuses;
  SendSQL("SELECT id, name 
              FROM attachstatuses JOIN attachstatusdefs 
              WHERE attachstatuses.statusid = attachstatusdefs.id 
              AND attach_id = $attachid");
  while (my ($id, $name) = FetchSQLData())
  {
    $statuses{$id} = $name;
  }

  # Create a list of valid statuses. !!!: Move this into versioncache?
  my @statusdefs;
  SendSQL("SELECT id, name FROM attachstatusdefs ORDER BY sortkey");
  while (my ($id, $name) = FetchSQLData())
  {
    push @statusdefs, { 'id' => $id , 'name' => $name };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'attachid' => $attachid , 
      'description' => $description , 
      'mimetype' => $mimetype , 
      'bug_id' => $bug_id , 
      'statuses' => \%statuses , 
      'statusdefs' => \@statusdefs , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };
  $template->process("attachment/console.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub adminDefs
{
  # Administer attachment status flags, which is the set of status flags 
  # that can be applied to an attachment.

  # If the user is seeing this screen as a result of doing something to
  # an attachment status flag, display a message about what happened
  # to that flag (i.e. "The attachment status flag was updated.").
  my ($message) = (@_);

  # Retrieve a list of attachment status flags and create an array of hashes
  # in which each hash contains the data for one flag.
  SendSQL("SELECT id, name, description, sortkey FROM attachstatusdefs ORDER BY sortkey");
  my @statusdefs;
  while (my ($id, $name, $description, $sortkey) = FetchSQLData())
  {
    push @statusdefs, { 'id' => $id , 'name' => $name , 'description' => $description , 'sortkey' => $sortkey };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'message' => $message , 
      'statusdefs' => \@statusdefs , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/adminDefs.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub editDef
{
  # Display a form for editing an existing attachment status flag.

  my ($id) = (@_);

  # Retrieve the definition from the database.
  SendSQL("SELECT name, description, sortkey FROM attachstatusdefs WHERE id = $id");
  my ($name, $desc, $sortkey) = FetchSQLData();

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'id' => $id , 
      'name' => $name , 
      'desc' => $desc , 
      'sortkey' => $sortkey , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/editDef.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub updateDef
{
  # Update an attachment status flag in the database.

  my ($id, $name, $desc, $sortkey) = (@_);

  # Quote the flag's name and description as appropriate for inclusion
  # in a SQL statement.
  $name = SqlQuote($name);
  $desc = SqlQuote($desc);

  SendSQL("LOCK TABLES attachstatusdefs WRITE");
  SendSQL("
           UPDATE  attachstatusdefs 
           SET     name = $name , 
                   description = $desc , 
                   sortkey = $sortkey 
           WHERE   id = $id
         ");
  SendSQL("UNLOCK TABLES");

  # Display the "administer attachment status flags" page
  # along with a message that the flag has been updated.
  adminDefs("The attachment status flag has been updated.");
}


sub createDef
{
  # Display a form for creating a new attachment status flag.

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("attachment/createDef.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub insertDef
{
  # Insert a new attachment status flag into the database.

  my ($name, $desc, $sortkey) = (@_);

  # Quote the flag's name and description as appropriate for inclusion
  # in a SQL statement.
  $name = SqlQuote($name);
  $desc = SqlQuote($desc);

  SendSQL("LOCK TABLES attachstatusdefs WRITE");
  SendSQL("
           INSERT INTO attachstatusdefs (name, description, sortkey)
           VALUES ($name, $desc, $sortkey)
          ");
  SendSQL("UNLOCK TABLES");

  # Display the "administer attachment status flags" page
  # along with a message that the flag has been created.
  adminDefs("The attachment status flag has been created.");
}


sub deleteDef
{
  # Delete an attachment status flag from the database.

  my ($id) = (@_);

  SendSQL("LOCK TABLES attachstatusdefs WRITE, attachstatuses WRITE");
  SendSQL("DELETE FROM attachstatuses WHERE statusid = $id");
  SendSQL("DELETE FROM attachstatusdefs WHERE id = $id");
  SendSQL("UNLOCK TABLES");

  # Display the "administer attachment status flags" page
  # along with a message that the flag has been deleted.
  adminDefs("The attachment status flag has been deleted.");
}


