#!/usr/bonsaitools/bin/perl -w
# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Contributor(s): Myk Melez <myk@mozilla.org>

############################################################################
# Script Initialization
############################################################################

# Make it harder for us to do dangerous things in Perl.
use diagnostics;
use strict;

# Include the Bugzilla CGI and general utility library.
require "CGI.pl";

# Establish a connection to the database backend.
ConnectToDatabase();

# The request module contains all the code for submitting and fulfilling 
# requests.  It is a member of the "Request" package, so its functions are
# referenced as "Request::functionName()".
use Request;

# Use the template toolkit (http://www.template-toolkit.org/) to generate
# the user interface (HTML pages and mail messages) using templates in the
# "templates/" subdirectory.
use Template;

# This is the global template object that gets used one or more times by
# the script when it needs to process a template and return the results.
# Configuration parameters can be specified here that apply to all templates
# processed in this file.
my $template = Template->new(
  {
    # Colon-separated list of directories containing templates.
    INCLUDE_PATH => 'templates' ,
    # Allow templates to be specified with relative paths.
    RELATIVE => 1 
  }
);

################################################################################
# Data Validation / Security Authorization
################################################################################

# Determine whether to use the action specified by the user or the default.
my $action = $::FORM{'action'} || 'create';

# !!! Validate all form fields!  
# For example: defid, tablename, fieldname, description, cclist, sortkey

################################################################################
# Main Body Execution
################################################################################

# All calls to this script should contain an "action" variable whose value
# determines what the user wants to do.  The code below checks the value of
# that variable and runs the appropriate code.

if ($action eq 'create')
{ 
  confirm_login();
  createRequest(); 
}

elsif ($action eq 'submit')  
{ 
  confirm_login();
  submitRequest(); 
}

elsif ($action eq 'queue') 
{ 
  displayQueue();
}

elsif ($action eq 'adminDefs') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  administerDefs();
}

elsif ($action eq 'createDef') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  createDef();
}

elsif ($action eq 'insertDef') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  insertDef(
    $::FORM{'objecttype'} , 
    $::FORM{'fieldname'} , 
    $::FORM{'fieldvalue'} , 
    $::FORM{'description'} , 
    $::FORM{'cclist'} , 
    $::FORM{'sortkey'} 
  );
}

elsif ($action eq 'editDef') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  editDef($::FORM{'defid'});
}

elsif ($action eq 'updateDef') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  updateDef(
    $::FORM{'defid'} , 
    $::FORM{'objecttype'} , 
    $::FORM{'fieldname'} , 
    $::FORM{'fieldvalue'} , 
    $::FORM{'description'} , 
    $::FORM{'cclist'} , 
    $::FORM{'sortkey'} 
  );
}

elsif ($action eq 'deleteDef') 
{ 
  confirm_login();
  UserInGroup("tweakparams")
    || DisplayError("You are not authorized to administer request definitions.")
    && exit;
  deleteDef($::FORM{'defid'});
}

else 
{ 
  # If the action that the user wants to take (specified in the "action" form field)
  # is none of the above listed actions, display an error telling the user that we
  # do not understand what they would like to do.
  DisplayError("I could not figure out what you wanted to do.")
    && exit;
}

################################################################################
# Functions
################################################################################

sub createRequest 
{
  my @defs;
  SendSQL("SELECT id, description FROM requestdefs 
           WHERE objecttype = '$::FORM{'objecttype'}' ORDER BY sortkey");
  while (my ($id, $description) = FetchSQLData())
  {
    push @defs , { 'id' => $id , 'description' => $description };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'defs' => \@defs , 
      'objecttype' => $::FORM{'objecttype'} , 
      'objectid' => $::FORM{'objectid'} , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/create.atml", $vars)
    || DisplayError("Template process failed: " . $template->error() . ".");
}


sub submitRequest 
{
  # Convert the contents of the web form into an array of hashes where each
  # hash represents a request and the array represents the list of requests.
  # If the "must verify fulfillers" flag is set, the user didn't provide
  # a string that uniquely identifies a fulfiller for one or more requests.
  my ($requests, $mustVerifyFulfillers) = convertFormFieldsToRequestObjects();

  if ($mustVerifyFulfillers)
  {
    # Display a form for verifying the fulfiller of each request.
    verifyFulfillers($requests);
  }
  else
  {
    # Get some information about the user as a hash.
    my $suppress_warning_that_var_is_used_only_once = $::userid;
    my $requester = Request::getUser($::userid);

    # Insert the requests into the database.
    insertRequestsIntoDatabase($requests, $requester);

    # Email the fulfillers to notify them of the requests.
    mailRequestsToFulfillers($requests, $requester);

    # Define the variables and functions that will be passed to the UI template.
    my ($bugsummary, $attachsummary, $bugid) = 
      Request::getSummaryInfo($::FORM{'objectid'}, $::FORM{'objecttype'});
    my $vars = 
      {
        'objectid'  => $::FORM{'objectid'} , 
        'objecttype' => $::FORM{'objecttype'} , 
        'requester' => $requester , 
        'requests'  => $requests , 
        'bugsummary' => $bugsummary , 
        'attachsummary' => $attachsummary , 
        'bugid' => $bugid , 
        'Param' => \&Param , # for retrieving global parameters
        'PerformSubsts' => \&PerformSubsts # for processing global parameters
      };

    # Return the appropriate HTTP response headers.
    print "Content-type: text/html\n\n";

    # Generate and return the UI (HTML page) from the appropriate template.
    $template->process("request/submitted.atml", $vars)
      || DisplayError("Template process failed: " . $template->error())
      && exit;
  }
}


sub displayQueue
{
  my @tables;
  my @criteria;
  
  if (defined $::FORM{'fulfiller'}) {
    my $str = SqlQuote($::FORM{'fulfiller'});
    my $op = $::FORM{'fulfillerop'} || "=";
    push @criteria , "(fulfillers.login_name $op $str OR fulfillers.realname $op $str)";
  }

  my $qry = "SELECT    fulfillers.realname , 
                       fulfillers.login_name , 
                       requesters.realname , 
                       requesters.login_name , 
                       requestdefs.description , 
                       requestdefs.objecttype , 
                       requests.objectid , 
                       requests.requestdate , 
                       requests.fulfilldate 
             FROM      requests, requestdefs, profiles AS requesters, profiles AS fulfillers
             WHERE     requests.defid = requestdefs.id
             AND       requests.requesterid = requesters.userid
             AND       requests.fulfillerid = fulfillers.userid";

  foreach my $criterion (@criteria)
  {
    $qry .= ' AND ' . $criterion . ' ';
  }

  $qry .= " AND requests.fulfilldate IS NULL ";

  $qry .= " ORDER BY  fulfillerid , fulfilldate , description , objectid";

  SendSQL($qry);
  my @recs;
  while (my ($fname, $femail, $rname, $remail, $desc, $otype, $oid, $rdate, $fdate) = FetchSQLData())
  {
    push 
      @recs , 
      { 
        'fulfillername' => $fname , 
        'fulfilleremail' => $femail , 
        'requestername' => $rname , 
        'requesteremail' => $remail , 
        'description' => $desc , 
        'objecttype' => $otype , 
        'objectid' => $oid , 
        'requesterdate' => $rdate , 
        'fulfillerdate' => $fdate 
      };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'recs' => \@recs , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/queue.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub administerDefs
{
  my ($message) = (@_);

  my @defs;
  SendSQL("SELECT id, objecttype, fieldname, fieldvalue, description, cclist, sortkey 
              FROM requestdefs ORDER BY sortkey");
  while (my ($id, $objecttype, $fieldname, $fieldvalue, $description, $cclist, $sortkey) = FetchSQLData())
  {
    push @defs, { 
                  'id' => $id , 
                  'objecttype' => $objecttype , 
                  'fieldname' => $fieldname , 
                  'fieldvalue' => $fieldvalue , 
                  'description' => $description , 
                  'cclist' => $cclist , 
                  'sortkey' => $sortkey 
                };
  }

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'message' => $message , 
      'defs' => \@defs , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/adminDefs.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub createDef
{
  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/createDef.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub insertDef
{
  my ($objecttype, $fieldname, $fieldvalue, $description, $cclist, $sortkey) = (@_);

  SendSQL("LOCK TABLES requestdefs WRITE");
  SendSQL("SELECT MAX(id) FROM requestdefs");
  my $id = FetchSQLData() + 1;
  SendSQL(
             "INSERT 
              INTO    requestdefs (id, objecttype, fieldname, fieldvalue, description, cclist, sortkey)
              VALUES  (" . 
                $id . ", " . 
                SqlQuote($objecttype) . ", " . 
                SqlQuote($fieldname) . ", " . 
                SqlQuote($fieldvalue) . ", " . 
                SqlQuote($description) . ", " . 
                SqlQuote($cclist) . ", " . 
                $sortkey . ")"
            );
  SendSQL("UNLOCK TABLES");

  administerDefs("The request definition has been created.");

}



sub editDef
{
  my ($id) = (@_);

  SendSQL("SELECT objecttype, fieldname, fieldvalue, description, cclist, sortkey 
              FROM requestdefs WHERE id = $id");
  my ($objecttype, $fieldname, $fieldvalue, $description, $cclist, $sortkey) = FetchSQLData();

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'id' => $id , 
      'objecttype' => $objecttype , 
      'fieldname' => $fieldname , 
      'fieldvalue' => $fieldvalue , 
      'description' => $description , 
      'cclist' => $cclist , 
      'sortkey' => $sortkey , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/editDef.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;

}


sub updateDef
{
  my ($id, $objecttype, $fieldname, $fieldvalue, $description, $cclist, $sortkey) = (@_);

  SendSQL("LOCK TABLES requestdefs WRITE");
  SendSQL("UPDATE  requestdefs 
              SET     
                      objecttype = " . SqlQuote($objecttype) . " , 
                      fieldname = " . SqlQuote($fieldname) . " , 
                      fieldvalue = " . SqlQuote($fieldvalue) . " , 
                      description = " . SqlQuote($description) . " , 
                      cclist = " . SqlQuote($cclist) . " , 
                      sortkey = $sortkey
             WHERE    id = $id
             ");
  SendSQL("UNLOCK TABLES");

  administerDefs("The request definition has been updated.");

}


sub deleteDef
{
  my ($id) = (@_);

  SendSQL("LOCK TABLES requestdefs WRITE, requests WRITE");
  SendSQL("DELETE FROM requests WHERE defid = $id");
  SendSQL("DELETE FROM requestdefs WHERE id = $id");
  SendSQL("UNLOCK TABLES");

  administerDefs("The request definition has been deleted.");

}


sub convertFormFieldsToRequestObjects
{
  # Use MFORM again so Perl does not report a warning
  my $neverused = $::MFORM{'defid'};

  # Flag for whether or not we must request verification of the fulfillers
  # (if the user did not uniquely identify them).
  my $mustVerifyFulfillers = 0;

  # Process the form data and create an array of requests.
  my @requests;
  for my $defid (@{$::MFORM{'defid'}})
  {
    # Only deal with requests for which the user entered a fulfiller,
    # since this is how the user specifies which requests they are making.
    $::FORM{"fulfillername-$defid"} || next;

    # To reduce the size of queries, require the user to enter at least 
    # three characters of each fulfiller's name.
    (length $::FORM{"fulfillername-$defid"} >= 3)
      || DisplayError("You must enter at least three characters of each fulfiller's name.")
      && exit;

    # Create the request object and populate it with data from the form.
    my %request;
    $request{'defid'} = $defid;
    $request{'comments'} = $::FORM{"comments-$defid"};
    $request{'fulfillername'} = $::FORM{"fulfillername-$defid"};

    # Get information about the request definition.
    SendSQL("SELECT description, cclist, fieldname, fieldvalue FROM requestdefs WHERE id = $defid");
    ($request{'description'}, $request{'cclist'}, $request{'fieldname'}, $request{'fieldvalue'}) = FetchSQLData();

    # Get a list of potential fulfillers whose login name (email address) 
    # or real name matches the substring entered by the user
    $request{'fulfillers'} = Request::matchUsers($::FORM{"fulfillername-$defid"});

    # If there is only one fulfiller match, make them the fulfiller.
    # Otherwise, set the "must verify fulfillers" flag.
    if ( scalar(@{$request{'fulfillers'}}) == 1 ) {
      $request{'fulfiller'} = $request{'fulfillers'}[0];
    } else {
      $mustVerifyFulfillers = 1;
    }

    # Add the request to the array of requests.
    push @requests , \%request;
  }

  # Return the list of requests along with whether or not the fulfillers
  # must be verified.
  return \@requests , $mustVerifyFulfillers;
}


sub verifyFulfillers 
{
  my ($requests) = @_;

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'objecttype' => $::FORM{'objecttype'} , 
      'objectid' => $::FORM{'objectid'} , 
      'requests' => $requests , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  # Return the appropriate HTTP response headers.
  print "Content-type: text/html\n\n";

  # Generate and return the UI (HTML page) from the appropriate template.
  $template->process("request/verify.atml", $vars)
    || DisplayError("Template process failed: " . $template->error())
    && exit;
}


sub insertRequestsIntoDatabase
{
  # Takes an array of requests and inserts them all into the database.

  my ($requests, $requester) = @_;

  SendSQL("LOCK TABLE requests WRITE");

  # Get the greatest value currently being used as a unique identifier.
  SendSQL("SELECT MAX(id) FROM requests");
  my ($id) = FetchSQLData();

  for my $request (@$requests)
  {
    # Quote the request comments for insertion into the database.
    my $requestcomments = SqlQuote($request->{'comments'});

    # Increment the unique identifier.
    ++$id;

    # Insert the record into the database
    SendSQL("
      INSERT INTO requests ( id , requesterid , requestdate , fulfillerid , defid , objectid , comments)
      VALUES ( $id , $requester->{'userid'} , NOW() , $request->{'fulfiller'}->{'userid'} , 
               $request->{'defid'} , $::FORM{'objectid'} , $requestcomments )
    ");
  }

  SendSQL("UNLOCK TABLES");

}


sub mailRequestsToFulfillers
{
  # Given an array of requests, notify the fulfillers by email.

  my ($requests, $requester) = @_;

  my $sendmailparam = Param("sendmailnow") ? "" : "-ODeliveryMode=deferred";

  my ($bugsummary, $attachsummary, $bugid) = Request::getSummaryInfo($::FORM{'objectid'}, $::FORM{'objecttype'});

  # Define the variables and functions that will be passed to the UI template.
  my $vars = 
    {
      'objectid'  => $::FORM{'objectid'} , 
      'objecttype'  => $::FORM{'objecttype'} , 
      'requests'  => $requests , 
      'requester' => $requester , 
      'bugsummary' => $bugsummary , 
      'attachsummary' => $attachsummary , 
      'bugid' => $bugid , 
      'Param' => \&Param , # for retrieving global parameters
      'PerformSubsts' => \&PerformSubsts # for processing global parameters
    };

  foreach my $request (@$requests)
  {
    $vars->{'request'} = $request;

    open(SENDMAIL, "|/usr/lib/sendmail $sendmailparam -t")
      || DisplayError("Can't open sendmail")
      && exit;

    $template->process("request/submitmail.atml", $vars, \*SENDMAIL)
      || DisplayError("Template process failed: " . $template->error())
      && exit;

    close SENDMAIL;
  }
}
